#include <alloc.h>
#include <dos.h>
#include <stdio.h>
#include <bios.h>
#include <fcntl.h>

#define D_READ	2
#define	D_WRITE	3

typedef struct {
  int fd; /* or drive, if cm set */
  int cm, hm, sm;
  int c, h, s;
} oread;

void *oread_open(char *fn)
{
  oread *or;
  if ((strlen(fn) == 2) && (fn[1] == ':'))
  {
    struct REGPACK r;
    char buf[512];
    int status;

    switch (fn[0])
    {
      case 'a':
      case 'A':
        or = (oread *)malloc(sizeof(oread));
        or->fd = 0;
        break;
      case 'b':
      case 'B':
        or = (oread *)malloc(sizeof(oread));
        or->fd = 1;
        break;
      default:
        printf("Invalid drive specified: %s\n", fn);
        return 0;
    }
    while ((status = biosdisk(D_READ, or->fd, 0, 0, 1, 1, buf)) == 6)
      /* wait for valid read */ ;
    if (status)
      return 0;

    r.r_ax = 0x0800;
    r.r_dx = or->fd;
    intr(0x13, &r);
    or->cm = (((r.r_cx & 0x00c0) << 2) | ((r.r_cx >> 8) & 0xff))+1;
    or->hm = (r.r_dx >> 8)+1;
    or->sm = r.r_cx & 0x3f;
    or->c = or->h = 0;
    or->s = -1;
    return or;
  }
  else
  {
    int fd;
    fd = _open(fn, O_RDONLY);
    if (fd < 0)
      return 0;
    or = (oread *)malloc(sizeof(oread));
    or->fd = fd;
    or->cm = 0;
    return or;
  }
}

static void advance(void *rv)
{
  oread *r = (oread *)rv;
  if (r->c == r->cm)
  {
    printf("Insert next floppy . . .");
    fflush(stdout);
    getch();
    r->c = 0;
  }
  if (++ r->s == r->sm)
  {
    r->s = 0;
    if (++ r->h == r->hm)
    {
      r->h = 0;
      r->c ++;
    }
  }
}

void oread_read(void *rv, void *buffer)
{
  oread *r = (oread *)rv;
  if (r->cm)
  {
    advance(r);
    if (biosdisk(D_READ, r->fd, r->h, r->c, r->s+1, 1, buffer))
      biosdisk(D_READ, r->fd, r->h, r->c, r->s+1, 1, buffer);
  }
  else
  {
    if (_read(r->fd, buffer, 512) != 512)
    {
      printf("Unexpected end-of-file\n");
      exit(1);
    }
  }
}

void oread_skip(void *rv, long skip_bytes)
{
  oread *r = (oread *)rv;
  if (r->cm)
  {
    while (skip_bytes > 0)
    {
      advance(r);
      skip_bytes -= 512;
    }
  }
  else
  {
    lseek(r->fd, skip_bytes, 1);
  }
}

void oread_close(void *rv)
{
  oread *r = (oread *)rv;
  if (r->cm == 0)
    _close(r->fd);
  free(r);
}
